@@ -73,7 +73,7 @@
  * in re-enabling affected modules.
  */
 
-#define CL_FLEVEL 129
+#define CL_FLEVEL 161
 #define CL_FLEVEL_DCONF CL_FLEVEL
 #define CL_FLEVEL_SIGTOOL CL_FLEVEL
 
@@ -146,7 +146,7 @@ extern uint8_t cli_always_gen_section_hash;
      (size_t)(sb) + (size_t)(sb_size) <= (size_t)(bb_size) && \
      (size_t)(sb) <= (size_t)(bb_size))
 
-#define CLI_MAX_ALLOCATION (182 * 1024 * 1024)
+#define CLI_MAX_ALLOCATION (1024 * 1024 * 1024)
 
 #ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h> /* for NAME_MAX */
@@ -175,37 +175,40 @@ typedef struct bitset_tag {
     unsigned long length;
 } bitset_t;
 
+typedef struct image_fuzzy_hash {
+    uint8_t hash[8];
+} image_fuzzy_hash_t;
+
 typedef struct recursion_level_tag {
     cli_file_t type;
     size_t size;
     cl_fmap_t *fmap;                      /* The fmap for this layer. This used to be in an array in the ctx. */
     uint32_t recursion_level_buffer;      /* Which buffer layer in scan recursion. */
     uint32_t recursion_level_buffer_fmap; /* Which fmap layer in this buffer. */
-    bool is_normalized_layer;             /* Indicates that the layer should be skipped when checking container and intermediate types. */
+    uint32_t attributes;                  /* layer attributes. */
+    image_fuzzy_hash_t image_fuzzy_hash;  /* Used for image/graphics files to store a fuzzy hash. */
+    bool calculated_image_fuzzy_hash;     /* Used for image/graphics files to store a fuzzy hash. */
 } recursion_level_t;
-// #define CONTAINER_FLAG_VALID 0x01
+
+typedef void *evidence_t;
 
 /* internal clamav context */
 typedef struct cli_ctx_tag {
-    char *target_filepath;    /**< (optional) The filepath of the original scan target. */
-    const char *sub_filepath; /**< (optional) The filepath of the current file being parsed. May be a temp file. */
-    char *sub_tmpdir;         /**< The directory to store tmp files at this recursion depth. */
-    const char **virname;
-    unsigned int num_viruses;
+    char *target_filepath;    /* (optional) The filepath of the original scan target. */
+    const char *sub_filepath; /* (optional) The filepath of the current file being parsed. May be a temp file. */
+    char *sub_tmpdir;         /* The directory to store tmp files at this recursion depth. */
+    evidence_t evidence;      /* Stores the evidence for this scan to alert (alerting indicators). */
     unsigned long int *scanned;
     const struct cli_matcher *root;
     const struct cl_engine *engine;
     uint64_t scansize;
     struct cl_scan_options *options;
     unsigned int scannedfiles;
-    unsigned int found_possibly_unwanted;
-    unsigned int corrupted_input;
-    unsigned int img_validate;
+    unsigned int corrupted_input;       /* Setting this flag will prevent the PE parser from reporting "broken executable" for unpacked/reconstructed files that may not be 100% to spec. */
     recursion_level_t *recursion_stack; /* Array of recursion levels used as a stack. */
     uint32_t recursion_stack_size;      /* stack size must == engine->max_recursion_level */
     uint32_t recursion_level;           /* Index into recursion_stack; current fmap recursion level from start of scan. */
     fmap_t *fmap;                       /* Pointer to current fmap in recursion_stack, varies with recursion depth. For convenience. */
-    bool next_layer_is_normalized;      /* Indicate that the next fmap pushed to the stack is normalized and should be ignored when checking container/intermediate types */
     unsigned char handlertype_hash[16];
     struct cli_dconf *dconf;
     bitset_t *hook_lsig_matches;
@@ -328,15 +331,15 @@ struct cl_engine {
     /* Limits */
     uint32_t maxscantime;         /* Time limit (in milliseconds) */
     uint64_t maxscansize;         /* during the scanning of archives this size
-				           * will never be exceeded
-				           */
+                                   * will never be exceeded
+                                   */
     uint64_t maxfilesize;         /* compressed files will only be decompressed
-				           * and scanned up to this size
-				           */
+                                   * and scanned up to this size
+                                   */
     uint32_t max_recursion_level; /* maximum recursion level for archives */
     uint32_t maxfiles;            /* maximum number of files to be scanned
-				           * within a single archive
-				           */
+                                   * within a single archive
+                                   */
     /* This is for structured data detection.  You can set the minimum
      * number of occurrences of an CC# or SSN before the system will
      * generate a notification.
@@ -353,15 +356,15 @@ struct cl_engine {
     struct cli_matcher *hm_mdb;
     /* hash matcher for MD5 sigs for PE import tables */
     struct cli_matcher *hm_imp;
-    /* hash matcher for whitelist db */
+    /* hash matcher for allow list db */
     struct cli_matcher *hm_fp;
 
     /* Container metadata */
     struct cli_cdb *cdb;
 
     /* Phishing .pdb and .wdb databases*/
-    struct regex_matcher *whitelist_matcher;
-    struct regex_matcher *domainlist_matcher;
+    struct regex_matcher *allow_list_matcher;
+    struct regex_matcher *domain_list_matcher;
     struct phishcheck *phishcheck;
 
     /* Dynamic configuration */
@@ -394,6 +397,9 @@ struct cl_engine {
     /* Database information from .info files */
     struct cli_dbinfo *dbinfo;
 
+    /* Signature counting, for progress callbacks */
+    size_t num_total_signatures;
+
     /* Used for memory pools */
     mpool_t *mempool;
 
@@ -401,6 +407,7 @@ struct cl_engine {
     crtmgr cmgr;
 
     /* Callback(s) */
+    clcb_file_inspection cb_file_inspection;
     clcb_pre_cache cb_pre_cache;
     clcb_pre_scan cb_pre_scan;
     clcb_post_scan cb_post_scan;
@@ -410,6 +417,12 @@ struct cl_engine {
     clcb_hash cb_hash;
     clcb_meta cb_meta;
     clcb_file_props cb_file_props;
+    clcb_progress cb_sigload_progress;
+    void *cb_sigload_progress_ctx;
+    clcb_progress cb_engine_compile_progress;
+    void *cb_engine_compile_progress_ctx;
+    clcb_progress cb_engine_free_progress;
+    void *cb_engine_free_progress_ctx;
 
     /* Used for bytecode */
     struct cli_all_bc bcs;
@@ -490,6 +503,12 @@ struct cl_settings {
     clcb_hash cb_hash;
     clcb_meta cb_meta;
     clcb_file_props cb_file_props;
+    clcb_progress cb_sigload_progress;
+    void *cb_sigload_progress_ctx;
+    clcb_progress cb_engine_compile_progress;
+    void *cb_engine_compile_progress_ctx;
+    clcb_progress cb_engine_free_progress;
+    void *cb_engine_free_progress_ctx;
 
     /* Engine max settings */
     uint64_t maxembeddedpe;      /* max size to scan MSEXE for PE */
@@ -680,8 +699,8 @@ static inline void cli_writeint32(void *offset, uint32_t value)
 /**
  * @brief Append an alert.
  *
- * An FP-check will verify that the file is not whitelisted.
- * The whitelist check does not happen before the scan because file whitelisting
+ * An FP-check will verify that the file is not allowed.
+ * The allow list check does not happen before the scan because allowing files
  * is so infrequent that such action would be detrimental to performance.
  *
  * TODO: Replace implementation with severity scale, and severity threshold
@@ -711,11 +730,23 @@ cl_error_t cli_append_virus(cli_ctx *ctx, const char *virname);
  * @param virname   The alert name.
  * @return cl_error_t CL_VIRUS if scan should be halted due to an alert, CL_CLEAN if scan should continue.
  */
-cl_error_t cli_append_possibly_unwanted(cli_ctx *ctx, const char *virname);
+cl_error_t cli_append_potentially_unwanted(cli_ctx *ctx, const char *virname);
+
+/**
+ * @brief If the SCAN_HEURISTIC_EXCEEDS_MAX option is enabled, append a "potentially unwanted" indicator.
+ *
+ * There is no return value because the caller should select the appropriate "CL_EMAX*" error code regardless
+ * of whether or not an FP sig is found, or allmatch is enabled, or whatever.
+ * That is, the scan must not continue because of an FP sig.
+ *
+ * @param ctx       The scan context.
+ * @param virname   The name of the potentially unwanted indicator.
+ */
+void cli_append_potentially_unwanted_if_heur_exceedsmax(cli_ctx *ctx, char *virname);
 
 const char *cli_get_last_virus(const cli_ctx *ctx);
 const char *cli_get_last_virus_str(const cli_ctx *ctx);
-void cli_virus_found_cb(cli_ctx *ctx);
+void cli_virus_found_cb(cli_ctx *ctx, const char *virname);
 
 /**
  * @brief Push a new fmap onto our scan recursion stack.
@@ -726,9 +757,10 @@ void cli_virus_found_cb(cli_ctx *ctx);
  * @param map           The fmap for the new layer.
  * @param type          The file type. May be CL_TYPE_ANY if unknown. Can change it later with cli_recursion_stack_change_type().
  * @param is_new_buffer true if the fmap represents a new buffer/file, and not some window into an existing fmap.
+ * @param attributes    Layer attributes for the thing to be scanned.
  * @return cl_error_t   CL_SUCCESS if successful, else CL_EMAXREC if exceeding the max recursion depth.
  */
-cl_error_t cli_recursion_stack_push(cli_ctx *ctx, cl_fmap_t *map, cli_file_t type, bool is_new_buffer);
+cl_error_t cli_recursion_stack_push(cli_ctx *ctx, cl_fmap_t *map, cli_file_t type, bool is_new_buffer, uint32_t attributes);
 
 /**
  * @brief Pop off a layer of our scan recursion stack.
@@ -818,6 +850,12 @@ void cli_infomsg(const cli_ctx *ctx, const char *fmt, ...) __attribute__((format
 void cli_infomsg(const cli_ctx *ctx, const char *fmt, ...);
 #endif
 
+#ifdef __GNUC__
+void cli_infomsg_simple(const char *fmt, ...) __attribute__((format(printf, 1, 2)));
+#else
+void cli_infomsg_simple(const char *fmt, ...);
+#endif
+
 void cli_logg_setup(const cli_ctx *ctx);
 void cli_logg_unsetup(void);
 
@@ -845,12 +883,22 @@ void cli_logg_unsetup(void);
 #define __hot__
 #endif
 
-#define cli_dbgmsg (!UNLIKELY(cli_debug_flag)) ? (void)0 : cli_dbgmsg_internal
+#ifdef __GNUC__
+inline void cli_dbgmsg(const char *str, ...) __attribute__((format(printf, 1, 2)));
+#else
+inline void cli_dbgmsg(const char *str, ...);
+#endif
 
 #ifdef __GNUC__
-void cli_dbgmsg_internal(const char *str, ...) __attribute__((format(printf, 1, 2)));
+void cli_dbgmsg_no_inline(const char *str, ...) __attribute__((format(printf, 1, 2)));
 #else
-void cli_dbgmsg_internal(const char *str, ...);
+void cli_dbgmsg_no_inline(const char *str, ...);
+#endif
+
+#ifdef __GNUC__
+size_t cli_eprintf(const char *str, ...) __attribute__((format(printf, 1, 2)));
+#else
+size_t cli_eprintf(const char *str, ...);
 #endif
 
 #ifdef HAVE_CLI_GETPAGESIZE
@@ -864,6 +912,7 @@ static inline int cli_getpagesize(void)
     GetSystemInfo(&si);
     return si.dwPageSize;
 }
+#define HAVE_CLI_GETPAGESIZE 1
 #else /* ! _WIN32 */
 #if HAVE_SYSCONF_SC_PAGESIZE
 static inline int cli_getpagesize(void)
@@ -884,13 +933,49 @@ static inline int cli_getpagesize(void)
 
 void *cli_malloc(size_t nmemb);
 void *cli_calloc(size_t nmemb, size_t size);
+
+/**
+ * @brief Wrapper around realloc that limits how much may be allocated to CLI_MAX_ALLOCATION.
+ *
+ * Please use CLI_REALLOC() with `goto done;` error handling instead.
+ *
+ * IMPORTANT: This differs from realloc() in that if size==0, it will NOT free the ptr.
+ *
+ * @param ptr
+ * @param size
+ * @return void*
+ */
 void *cli_realloc(void *ptr, size_t size);
+
+/**
+ * @brief Wrapper around realloc that limits how much may be allocated to CLI_MAX_ALLOCATION.
+ *
+ * Please use CLI_REALLOC() with `goto done;` error handling instead.
+ *
+ * IMPORTANT: This differs from realloc() in that if size==0, it will NOT free the ptr.
+ *
+ * WARNING: This differs from cli_realloc() in that it will free the ptr if the allocation fails.
+ * If you're using `goto done;` error handling, this may result in a double-free!!
+ *
+ * @param ptr
+ * @param size
+ * @return void*
+ */
 void *cli_realloc2(void *ptr, size_t size);
+
 char *cli_strdup(const char *s);
 int cli_rmdirs(const char *dirname);
 char *cli_hashstream(FILE *fs, unsigned char *digcpy, int type);
 char *cli_hashfile(const char *filename, int type);
-int cli_unlink(const char *pathname);
+
+/**
+ * @brief unlink() with error checking
+ *
+ * @param pathname the file path to unlink
+ * @return cl_error_t CL_SUCCESS if successful, CL_EUNLINK if unlink() failed
+ */
+cl_error_t cli_unlink(const char *pathname);
+
 size_t cli_readn(int fd, void *buff, size_t count);
 size_t cli_writen(int fd, const void *buff, size_t count);
 const char *cli_gettmpdir(void);
@@ -936,7 +1021,7 @@ char *cli_newfilepath(const char *dir, const char *fname);
  * If the dir is not provided, the engine->tmpdir will be used.
  *
  * @param dir        Alternative temp directory (optional).
- * @param prefix  	 (Optional) Base filename for new file.
+ * @param fname  	 Filename for new file.
  * @param[out] name  Allocated filepath, must be freed by caller.
  * @param[out] fd    File descriptor of open temp file.
  */
@@ -948,6 +1033,7 @@ cl_error_t cli_newfilepathfd(const char *dir, char *fname, char **name, int *fd)
  * Caller is responsible for freeing the filename.
  *
  * @param dir 	 Alternative temp directory. (optional)
+ * @param prefix (Optional) Prefix for new file tempfile.
  * @return char* filename or NULL.
  */
 char *cli_gentemp_with_prefix(const char *dir, const char *prefix);
@@ -981,7 +1067,7 @@ cl_error_t cli_gentempfd(const char *dir, char **name, int *fd);
  * @param[out] fd    File descriptor of open temp file.
  * @return cl_error_t CL_SUCCESS, CL_ECREAT, or CL_EMEM.
  */
-cl_error_t cli_gentempfd_with_prefix(const char *dir, char *prefix, char **name, int *fd);
+cl_error_t cli_gentempfd_with_prefix(const char *dir, const char *prefix, char **name, int *fd);
 
 unsigned int cli_rndnum(unsigned int max);
 int cli_filecopy(const char *src, const char *dest);
@@ -990,8 +1076,8 @@ void cli_bitset_free(bitset_t *bs);
 int cli_bitset_set(bitset_t *bs, unsigned long bit_offset);
 int cli_bitset_test(bitset_t *bs, unsigned long bit_offset);
 const char *cli_ctime(const time_t *timep, char *buf, const size_t bufsize);
-void cli_append_virus_if_heur_exceedsmax(cli_ctx *, char *);
-cl_error_t cli_checklimits(const char *, cli_ctx *, unsigned long, unsigned long, unsigned long);
+
+cl_error_t cli_checklimits(const char *who, cli_ctx *ctx, unsigned long need1, unsigned long need2, unsigned long need3);
 
 /**
  * @brief Call before scanning a file to determine if we should scan it, skip it, or abort the entire scanning process.
@@ -1093,6 +1179,19 @@ cl_error_t cli_ftw(char *base, int flags, int maxdepth, cli_ftw_cb callback, str
 
 const char *cli_strerror(int errnum, char *buf, size_t len);
 
+#ifdef _WIN32
+/**
+ * @brief   Attempt to get a filename from an open file handle.
+ *
+ * Windows only.
+ *
+ * @param hFile          File handle
+ * @param[out] filepath  Will be set to file path if found, or NULL.
+ * @return cl_error_t    CL_SUCCESS if found, else an error code.
+ */
+cl_error_t cli_get_filepath_from_handle(HANDLE hFile, char **filepath);
+#endif
+
 /**
  * @brief   Attempt to get a filename from an open file descriptor.
  *
@@ -1114,17 +1213,166 @@ cl_error_t cli_get_filepath_from_filedesc(int desc, char **filepath);
  * to get the real path.
  *
  * @param desc          A file path to evaluate.
- * @param char*         [out] A malloced string containing the real path.
+ * @param[out] char*    A malloced string containing the real path.
  * @return cl_error_t   CL_SUCCESS if found, else an error code.
  */
 cl_error_t cli_realpath(const char *file_name, char **real_filename);
 
+/**
+ * @brief   Get the libclamav debug flag (e.g. if debug logging is enabled)
+ *
+ * This is required for unit tests to be able to link with clamav.dll and not
+ * directly manipulate libclamav global variables.
+ */
+uint8_t cli_get_debug_flag(void);
+
+/**
+ * @brief   Set the libclamav debug flag to a specific value.
+ *
+ * The public cl_debug() API will only ever enable debug mode, it won't disable debug mode.
+ *
+ * This is required for unit tests to be able to link with clamav.dll and not
+ * directly manipulate libclamav global variables.
+ */
+uint8_t cli_set_debug_flag(uint8_t debug_flag);
+
+#ifndef STRDUP
+#define STRDUP(buf, var, ...) \
+    do {                      \
+        var = strdup(buf);    \
+        if (NULL == var) {    \
+            do {              \
+                __VA_ARGS__;  \
+            } while (0);      \
+            goto done;        \
+        }                     \
+    } while (0)
+#endif
+
+#ifndef CLI_STRDUP
+#define CLI_STRDUP(buf, var, ...) \
+    do {                          \
+        var = cli_strdup(buf);    \
+        if (NULL == var) {        \
+            do {                  \
+                __VA_ARGS__;      \
+            } while (0);          \
+            goto done;            \
+        }                         \
+    } while (0)
+#endif
+
 #ifndef FREE
-#define FREE(VAR)      \
-    if (NULL != VAR) { \
-        free(VAR);     \
-        VAR = NULL;    \
-    }
-#endif /* FREE */
+#define FREE(var)          \
+    do {                   \
+        if (NULL != var) { \
+            free(var);     \
+            var = NULL;    \
+        }                  \
+    } while (0)
+#endif
+
+#ifndef MALLOC
+#define MALLOC(var, size, ...) \
+    do {                       \
+        var = malloc(size);    \
+        if (NULL == var) {     \
+            do {               \
+                __VA_ARGS__;   \
+            } while (0);       \
+            goto done;         \
+        }                      \
+    } while (0)
+#endif
+
+#ifndef CLI_MALLOC
+#define CLI_MALLOC(var, size, ...) \
+    do {                           \
+        var = cli_malloc(size);    \
+        if (NULL == var) {         \
+            do {                   \
+                __VA_ARGS__;       \
+            } while (0);           \
+            goto done;             \
+        }                          \
+    } while (0)
+#endif
+
+#ifndef CALLOC
+#define CALLOC(var, nmemb, size, ...) \
+    do {                              \
+        (var) = calloc(nmemb, size);  \
+        if (NULL == var) {            \
+            do {                      \
+                __VA_ARGS__;          \
+            } while (0);              \
+            goto done;                \
+        }                             \
+    } while (0)
+#endif
+
+#ifndef CLI_CALLOC
+#define CLI_CALLOC(var, nmemb, size, ...) \
+    do {                                  \
+        (var) = cli_calloc(nmemb, size);  \
+        if (NULL == var) {                \
+            do {                          \
+                __VA_ARGS__;              \
+            } while (0);                  \
+            goto done;                    \
+        }                                 \
+    } while (0)
+#endif
+
+#ifndef VERIFY_POINTER
+#define VERIFY_POINTER(ptr, ...) \
+    do {                         \
+        if (NULL == ptr) {       \
+            do {                 \
+                __VA_ARGS__;     \
+            } while (0);         \
+            goto done;           \
+        }                        \
+    } while (0)
+#endif
+
+/**
+ * @brief Wrapper around realloc that limits how much may be allocated to CLI_MAX_ALLOCATION.
+ *
+ * IMPORTANT: This differs from realloc() in that if size==0, it will NOT free the ptr.
+ *
+ * NOTE: cli_realloc() will NOT free ptr if size==0. It is safe to free ptr after `done:`.
+ *
+ * @param ptr
+ * @param size
+ * @return void*
+ */
+#ifndef CLI_REALLOC
+#define CLI_REALLOC(ptr, size, ...)          \
+    do {                                     \
+        void *vTmp = cli_realloc(ptr, size); \
+        if (NULL == vTmp) {                  \
+            do {                             \
+                __VA_ARGS__;                 \
+            } while (0);                     \
+            goto done;                       \
+        }                                    \
+        ptr = vTmp;                          \
+    } while (0)
+#endif
+
+/*This is a duplicate from other PR's.*/
+#ifndef CLI_STRDUP
+#define CLI_STRDUP(buf, var, ...) \
+    do {                          \
+        var = cli_strdup(buf);    \
+        if (NULL == var) {        \
+            do {                  \
+                __VA_ARGS__;      \
+            } while (0);          \
+            goto done;            \
+        }                         \
+    } while (0)
+#endif
 
 #endif
